#!/usr/bin/perl


use strict;

my %ExistLink;
my $process;
my $View;
my $AccumulateFlag;

#skyrails
my @nodes;
my @ties;


sub cleanstring {
  my $name = join('_',@_);
  #my $name = lc(join('_',@_));
  #$name =~ s/[\}\{} \+\)\(\.\-\\\$\:\&]/_/g;
  $name =~ s/^\s+//;
  $name =~ s/\s+$//;
  $name =~ s/'//g;
  $name =~ s/\\/\\\\/g;
#  $name = "\"$name\"";
return $name;
}

sub ParseCellPutGet {
  my ($Source, $line) = @_;
  #add zeroout and import view names in label if available
  my $edge = "";
  #add import view name
  $edge = ", label=$View" if ($View ne "");
  #normal = store, bold = accumulate
  $edge = "$edge, style=bold, arrowhead=dot" if ($AccumulateFlag);
  
  #exit parsing if CellPutGet function commented out
  return if (/\#.*Cell/);
  if (/Cell(Put|Get)(N|S)\(([^,]*),([^,]*)/i) {
      my $flow = $1;
      my $c3 = cleanstring($3);
      my $c4 = cleanstring($4);

      if ($line =~ /CellPut.\((.*)$/i) {
        #parse CellPutArgs
        my $CellPutArgs = $1;
        my $cube = "";
        my $nested = 0;
        my $reached = 0;

        #1 character at a time
        while ($CellPutArgs =~ /(.)/g) {
          my $c = $1;
          $nested++ if $c eq "(";
          $nested-- if $c eq ")";
          #if we reached the 2nd argument, 
          if($reached) {
          #keep filling the cube name until next comma
            if($c eq ",") {
              $reached--;
            } else {
              $cube = "$cube$1";
            }
          } else {
          #if we get an unnested comma then we reached the 2nd argument i.e cube name
            $reached++ if ($c eq "," && $nested == 0 && $cube eq "");
          }
        }
        #print STDERR "$cube\n";
        $c4 = cleanstring($cube);
      }

      if (($flow =~ /put/i) && ($ExistLink{"$process$c4"} ne "yes")) {
        $ExistLink{"$process$c4"} = "yes";
        #create a link: process -> cube
        push(@ties, "\"$process\" \"$c4\" 1 0\n") if  ($c4 ne "");
        #print FLR "$process -> $c4 [color=darkturquoise$edge];\n" if ($c4 ne "");
        if ($ExistLink{"$Source$process"} ne "yes") {
        #create a link: source -> process -> cube
          $ExistLink{"$Source$process"} = "yes";
          push(@ties, "\"$Source\" \"$process\" 1 0\n") if  ($Source ne "");
          #print FLR "$Source -> $process [color=red$edge];\n" if ($Source ne "");
        }
      }

      if (($flow =~ /get/i) && ($ExistLink{"$c3$process"} ne "yes")) {
        $ExistLink{"$c3$process"} = "yes";
        push(@ties, "\"$c3\" \"$process\" 1 0\n") if  ($c3 ne "");
        #print FLR "$c3 -> $process [color=blue$edge];\n" if ($c3 ne "");
      }      
  }
}

print "Enter the path to your TM1 server data folder: ";

my $dirname = <STDIN>;
chomp($dirname);

#my $dirname = $ARGV[0];
if ($dirname eq "") {
    print "TM1 to skyrails input data file generator beta\n";
    print "Usage: TM1skyrails.pl\n";
    print "run it from a DOS window for debug messages";
    exit;
}

print "Generating input and launching skyrails. Please wait...\n\n";
print "Once the skyrails window appears, click the folder icon\n";
print "a new menu line will appear and then click on TM1";

opendir THISDIR, $dirname or die "Cannot open $dirname $!";
while (my $thisfile = readdir THISDIR) {
    my $DataSource = "";
    my $SourceType = "";
    open(FLH, "$dirname/$thisfile");
    #rules files
    if ($thisfile =~ /(.*)\.rux$/i) {
        #print "#$dirname/$thisfile\n";
        my $cube = cleanstring($1);
        while (<FLH>) {
            #inter-cubes flow
            if(/.*=>\s*DB\(\'([^\']*)\'/) {
                my $destcube = cleanstring($1);
                #test if link already added in graph
                my $link = "${cube}${destcube}rulefeed";
                push(@ties, "\"$cube\" \"$destcube\" 0 1\n") if  ($ExistLink{$link} ne "yes");
                #print FLR "$cube -> $destcube [color=green];\n" if  ($ExistLink{$link} ne "yes");
                $ExistLink{$link} = "yes";
                
            }
            if(/.*=[^>]*DB\(\'([^\']*)\'/) {
              #loop as DB calls could be nested
              foreach my $sourcecube (/DB\(\'([^\']*)\'/) {
                  $sourcecube = cleanstring($sourcecube);
                  push(@ties, "\"$sourcecube\" \"$cube\" 0 1\n") if ($ExistLink{"${sourcecube}${cube}rule"} ne "yes");
                  #print FLR "$sourcecube -> $cube [color=darkgreen];\n" if ($ExistLink{"${sourcecube}${cube}rule"} ne "yes");
                  $ExistLink{"${sourcecube}${cube}rule"} = "yes";
              }
            }
        }
    }
    if ($thisfile =~ /(.*)\.cub$/i) {
      my $cube = $1;
      if ($ExistLink{$cube} ne "shape") {
          $ExistLink{$cube} = "shape";
          push(@nodes, "\"$cube\" \"cube\"\n");
      }
    }
    if ($thisfile =~ /(.*)\.pro$/i) {
        my $SourceType = "";
        my $Source = "";
        $AccumulateFlag = 0;
        $View = "";
        $process = cleanstring($1);
        
        #setup color and shape of processes
        if ($ExistLink{$process} ne "shape") {
          $ExistLink{$process} = "shape";
          push(@nodes, "\"$process\" \"process\"\n");
          #print FLR "$process [shape = record, color=blue];\n";
        }
        
        while (<FLH>) {
            #replace $cube with SQL statement,ascii file import, TM1 view
            #add print for special shape
            #datasource type

            if (/^562,"([A-Z]+)"/) {
              $SourceType = $1;
            }
            if (/^586,"(.*)"/) {
              #ASCII or VIEW source
              $Source = cleanstring($1);
              #same paths/files may be captilalised differently so get all lower cases
              $Source = lc($Source) if ($SourceType ne "VIEW");
              if ($ExistLink{$Source} ne "shape") {
                $ExistLink{$Source} = "shape";

                  push(@nodes, "\"$Source\" \"cube\"\n") if ($SourceType eq "VIEW");
                  push(@nodes, "\"$Source\" \"source\"\n") if ($SourceType eq "ODBC"); 
                  push(@nodes, "\"$Source\" \"file\"\n") if ($SourceType eq "CHARACTERDELIMITED"); 

                #my $shape = "ellipse";
                #$shape = "egg" if ($SourceType eq "ODBC");
                #$shape = "parallelogram" if ($SourceType eq "CHARACTERDELIMITED");
                #print FLR "$Source [color=red, shape=$shape];\n"
              }
            }
            if ((/^570,(.*)/) && ($SourceType eq "VIEW")) {
              #VIEW name
              $View = cleanstring($1);
            }
            #check for preset ZeroOut views
            if (/ZeroOut=1.ViewName=(.*).CubeLogChanges/) {
              my $ZeroOut = $1;
              my $labelprocess =$process;
              $labelprocess =~ s/^\"//;
              $labelprocess =~ s/\"$//;
              push(@nodes, "\"$process\" \"process\"\n" );
              #print FLR "$process [shape = record, color=blue, label=\"$labelprocess | 0out: $ZeroOut\"];\n";
            }
            #1503 = store (overwrite cell content with new value), 1504 = accumulate values
            #won't display because it comes after all the CellPut commands
            $AccumulateFlag = 1 if (/DataAction=1504/);
            
            ParseCellPutGet($Source, $_);
         }
    }
    if ($thisfile =~ /(.*)\.cub$/i) {
        my $cube = cleanstring($1);
        #skip control cubes
        #print "$cube;\n" if ($cube !~ /^_/);
    }
    close FLH;
}

#create input file
my $last = "";
my $id = 1;
open(FLR, "> data/TM1.vna") or die "Cannot create output file TM1.vna!\n";
print FLR "*Node Data\n\"ID\" \"fullname\" \"TM1type\"\n";
my @sortnodes = sort {$a cmp $b} @nodes;
foreach (@sortnodes) {
  my @ident = split(/\" \"/,$_);
  print FLR "$ident[0]\" $_" if ($last ne $_);
  $id = $id + 1 if ($last ne $_);
  $last = $_;
}

print FLR "*Tie Data\nFROM TO \"process\" \"rule\"\n";
my @sortties = sort {$a cmp $b} @ties;
foreach (@sortties) {
  print FLR "$_" if ($last ne $_);
  $last = $_;
}

close FLR;

#run skyrails
system("raex.exe");
